import hyperdiv as hd
import inspect
from button import button_examples
from checkbox import checkbox_examples
from box import box_examples
from hyperdiv.core.element import collect_static_props, Element
from hyperdiv.core.context import paused_collector

menu_items = [
    dict(
        title="Intro",
        children=[
            dict(title="Overview", route="/overview"),
            dict(title="Installation", route="install"),
            dict(title="Getting Started", route="/getting_started"),
        ],
    ),
    dict(
        title="Components",
        children=[
            dict(title="Box", route="/box"),
            dict(title="Button", route="/button"),
            dict(title="Checkbox", route="/checkbox"),
        ],
    ),
    dict(title="Prop Types", hide=True, route="/proptypes"),
]


class MenuRouter:
    def __init__(self, menu, default=None):
        self.name_map = dict()
        self.path_map = dict()
        self._init_maps(menu)
        self.location = hd.location()

        self.menu = menu
        self.active = self._get_active(default=default)

    def _init_maps(self, menu):
        for item in menu:
            if "route" in item:
                path = item["route"]
                name = item["title"]
                self.name_map[name] = path
                self.path_map[path] = name
            if "children" in item:
                children = item["children"]
                self._init_maps(children)

    def is_active(self, name):
        return self.active == name

    def set_active(self, name):
        if name not in self.name_map:
            raise Exception(f"Unknown route '{name}'.")
        self.active = name
        self.location.path = self.name_map[name]

    def _get_active(self, default=None):
        if self.location.path in self.path_map:
            return self.path_map[self.location.path]
        elif self.location.path == "/" and default and default in self.name_map:
            self.location.path = self.name_map[default]
            return default
        else:
            return None


def get_code(function):
    code = inspect.getsource(function)
    return "\n".join(line[4:] for line in code.split("\n")[1:]).strip()


def code_block(function):
    code = get_code(function)
    hd.code(
        code,
        language="python",
        font_size="small",
    )


def content_heading(title, description):
    hd.text(f"# {title}")
    hd.text(description, font_color=hd.lighten("font_color", 150))


def prop_table(props):
    cell_style = dict(padding=8)
    with hd.table():
        with hd.table_head():
            with hd.table_row():
                hd.table_cell("Name")
                hd.table_cell("Type")
                hd.table_cell("Default")
                hd.table_cell("Description")
        with hd.table_body():
            for (propname, prop) in props:
                with hd.scope(f"{propname}"):
                    with hd.table_row():
                        hd.table_cell(propname, **cell_style)
                        proptype = type(prop).__name__
                        hd.table_cell(
                            f"[{proptype}](./proptypes/{proptype})", **cell_style
                        )
                        value = prop.value
                        if callable(prop.value):
                            value = "function"
                        hd.table_cell(f"`{value}`", **cell_style)
                        hd.table_cell("", **cell_style)


def top_bar():
    """The top navigation bar.
    Currently nothing up here but the app title."""
    with hd.hbox(
        width="100%",
        padding=16,
        gap=16,
        align="center",
        height=48,
        border=(None, None, hd.theme().input_border, None),
    ):
        hd.text("**Hyperdiv documentation**")


def tree_menu(items, state):
    for item in items:
        title = item["title"]
        with hd.scope(title):
            if "children" in item and len(item["children"]) > 0:
                hd.text(f"### {title}")
                with hd.box(padding=[4, 0, 16, 0]):
                    tree_menu(item["children"], state)

            else:
                if state.selected == title:
                    bgcolor = hd.highlight(hd.theme().background_color, 25)
                else:
                    bgcolor = None

                if "hide" not in item or not item["hide"]:
                    item = hd.text(
                        title, background_color=bgcolor, padding=8, cursor="pointer"
                    )
                    if item.clicked:
                        state.selected = title


def side_menu(router):
    """The lefthand menu with the list of components.
    When elements are selected the router updates
    and corresponding content is rendered."""
    with hd.box(
        height="100%",
        shrink=0,
        width=200,
        background_color=hd.theme().surface_color,
        padding=8,
    ):
        menu_state = hd.state(selected=None)
        tree_menu(menu_items, menu_state)
        if menu_state.selected is None:
            menu_state.selected = router.active

    if menu_state.selected:
        router.set_active(menu_state.selected)


def examples(element_function_name):
    """Gets the list of examples to demonstrate for the component.
    Examples for a component are are currently defined in the global namespace
    And each is a list of list of tuples [(name, function), ...]
    e.g. each entry has the name of each example and the function that executes it.
    """
    examples = globals()[f"{element_function_name}_examples"]
    # Examples are a list of tuples [(name, function), ...]
    # The name of each example and the function that executes it.
    for example in examples:
        example_name = example[0]
        example_fn = example[1]
        with hd.scope(example_name):
            # Show each example code next to its code block
            with hd.box(gap=4):
                if len(example_name) > 0:
                    hd.text(f"#### {example_name}")
                with hd.hbox(gap=16):
                    with hd.box(width="50%"):
                        code_block(example_fn)
                    with hd.box(
                        background_color=hd.theme().surface_color,
                        padding=16,
                        gap=8,
                        width="50%",
                    ):
                        example_fn()


def organize_props_by_class(element_or_combo, element_name):
    by_class = dict()

    # If it's a standard element, collect the static props
    if inspect.isclass(element_or_combo) and issubclass(element_or_combo, Element):
        props = collect_static_props(element_or_combo)

    # Otherwise create an instance and get the props dynamically
    else:
        props = dict()
        with paused_collector():
            combo_instance = element_or_combo()
            unorganized_props = combo_instance._props
            for propname, prop in unorganized_props.items():
                static_props = collect_static_props(prop.parent_element.__class__)
                props[propname] = (prop, static_props[propname][1])

    # Store props by their source class
    for propname, (prop, propclass) in props.items():
        if propclass not in by_class:
            by_class[propclass] = []

        by_class[propclass].append((propname, prop))

    sorted_by_class = []
    for propclass, props in by_class.items():
        props = sorted(props, key=lambda p: p[0])
        sorted_by_class.append((propclass, props))

    # Sort the props in alphabetical order by class name,
    # But put the props defined directly by the element first.
    sorted_by_class = sorted(
        sorted_by_class,
        key=lambda p: "" if p[0].__name__ == element_name else p[0].__name__,
    )
    return sorted_by_class


def prop_tables(element_function_name):
    """Renders multiple tables, each representing props for an element.
    Props are defined as a list of tuples, [group_name, [props]]
    Where the first entry is the name of the group
    (e.g. "Props inherited from StyledElement", Or just an empy string
    if they are the default props.)
    Each prop is a dict with name, type, default, and description attributes."""

    element = getattr(hd, element_function_name)
    with hd.box(gap=24):
        props_by_class = organize_props_by_class(element, element_function_name)

        for propclass, props in props_by_class:
            with hd.scope(f"{element_function_name} {propclass}"):
                with hd.box(gap=4):
                    if propclass.__name__ != element_function_name:
                        hd.text(
                            f"#### Inherited from [{propclass.__name__}](./{propclass.__name__})"
                        )
                    prop_table(props)


def component_page(route):
    with hd.box():
        content_heading(
            route,
            f"""API documentation for the {route} component.
            Learn about the available props and the API.""",
        )

    element_function_name = route.lower().replace(" ", "_")
    with hd.box(gap=16):
        hd.text("### Examples")
        examples(element_function_name)

    with hd.box(gap=4):
        hd.text("### Props")
        prop_tables(element_function_name)


def get_content(route):
    components = None
    for item in menu_items:
        if item["title"] == "Components":
            components = item

    if route in [item["title"] for item in components["children"]]:
        component_page(route)
    elif route == "Overview":
        overview()
    elif route == "Installation":
        installation()
    elif route == "Getting Started":
        getting_started()


def overview():
    hd.text(
        """
        # Welcome to Hyperdiv
        Hyperdiv is an open-source Python-based UI framework
        for making interactive websites. No frontend experience is required.
        Everything can be built using basic Python.


        It includes a collection of prebuilt components that are ready to use
        right out of the box. The components can be composed together in arbitrary
        layouts, and stylized to achieve your own look and feel.

        Hyperdiv is simple by design and easy to connect with your data.
        It also includes a variety of options that make it easy to implement
        new reusable components.

        ## Advantages of Hyperdiv
        * **No frontend experience required.** There is no need to learn HTML,
        CSS, Javascript, React, or any other frontend library. Hyperdiv apps
        are written as regular functions in Python. Python's reactive framework
        also means that there are no confusing callbacks to handle events.
        * **Build internal tools faster.** 
        Hyperdiv is ideal for building internal tools, where you need something
        quick and dirty for the internals of your project.
        You might have data already set up, and so it's easy to just import
        that data right in Python and build your website as if you're building
        a standard backend CLI.
        * **Looks good by default.** We've designed Hyperdiv to look good out
        of the box so that you don't need to fuss around with styles and themes.
        You can if you'd like, but hyperdiv apps should be simple and usable
        from the start.
        """
    )


def installation():
    hd.text("# Installation")


def getting_started():
    hd.text("# Here's how to get started.")
    hd.text("""This section should probably have code for running a basic app.""")


def proptypes():
    hd.text("# Prop Types")
    proptypes = filter(lambda p: p.endswith("Prop"), dir(hd.elements.props))
    for proptype in proptypes:
        with hd.scope(proptype):
            with hd.box(gap=4):
                hd.text(f"#### {proptype}")
                proptype_class = getattr(hd.elements.props, proptype)
                hd.text(proptype_class.__doc__)


def main():
    hd.set_title("Hyperdiv Documentation")
    router = MenuRouter(menu_items, default="Overview")
    top_bar()

    # TODO: I'm not sure why I needed to set this hbox's vertical_scroll to True
    # But otherwise the entire page would scroll.
    with hd.hbox(height="100%", vertical_scroll=True):
        side_menu(router)
        # Use the scope to ensure each page has it's own
        # Scrollable container.
        with hd.scope(router.active):
            with hd.box(
                grow=1,
                gap=36,
                padding=[16, 40, 16, 40],
                vertical_scroll=True,
            ):
                if router.active is None:
                    if router.location.path.startswith("/proptypes/"):
                        proptypes()
                    else:
                        hd.text("#404 Not Found.")
                else:
                    get_content(router.active)


if __name__ == "__main__":
    hd.run(main)
