from ..core.element import Prop
from .prop_utils import parse_color, is_non_negative_number, number_to_str


class NativeProp(Prop):
    def __init__(self, type_, *args, **kwargs):
        self.type_ = type_
        super().__init__(*args, **kwargs)

    def parse(self, value):
        if not isinstance(value, self.type_):
            raise ValueError(
                f"Value of type `{self.type_.__name__}` expected "
                f"but got value `{value}` of type `{type(value).__name__}`."
            )
        return value


class IntProp(NativeProp):
    def __init__(self, value, **kwargs):
        super().__init__(int, value, **kwargs)


class BoolProp(NativeProp):
    def __init__(self, value, **kwargs):
        super().__init__(bool, value, **kwargs)


class StrProp(NativeProp):
    def __init__(self, value, **kwargs):
        super().__init__(str, value, **kwargs)


class PositiveIntProp(IntProp):
    def parse(self, value):
        value = super().parse(value)
        if value < 0:
            raise ValueError(f"A positive int expected but got {value}")
        return value


class OptionalStrProp(StrProp):
    def parse(self, value):
        if value is None:
            return value
        return super().parse(value)


class ListProp(Prop):
    def parse(self, value):
        if type(value) not in (list, tuple):
            raise ValueError("Value must be a list or a tuple.")
        return value


class TypedListProp(ListProp):
    def __init__(self, type_, value, **kwargs):
        self.type_ = type_
        super().__init__(value, **kwargs)

    def parse(self, value):
        return [self.type_(elem) for elem in super().parse(value)]


class OptionProp(Prop):
    def __init__(self, value, options, **kwargs):
        self.options = options
        super().__init__(value, **kwargs)

    def parse(self, value):
        if value not in self.options:
            options = ", ".join([str(o) for o in self.options])
            raise ValueError(f"Value {value} is not one of {options}")
        return value


class FontWeightProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, (None, "normal", "bold"), **kwargs)


class FontSizeProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(
            value,
            (
                None,
                "xx-small",
                "x-small",
                "small",
                "medium",
                "large",
                "x-large",
                "xx-large",
                "xxx-large",
            ),
            **kwargs,
        )


class FontsProp(Prop):
    def parse(self, fonts):
        if fonts is None:
            return None

        # We only send webfonts that with urls
        webfonts = []
        for url in fonts.values():
            if url is not None and url not in webfonts:
                webfonts.append(url)
        return webfonts


class FontFamilyProp(Prop):
    def parse(self, values):
        if values is None:
            return None

        # Checks if the font is included
        if type(values) in (list, tuple):
            return ", ".join(values)

        return values


class SizeProp(Prop):
    """
    Accepts the following types:

    1. None
    2. A number: int or float
    3. A percent string "x%", where 0 <= x <= 100
    """

    def parse(self, size):
        def is_percent(value):
            if type(value) is str and value.endswith("%"):
                num = float(value[:-1])
                return num >= 0 and num <= 100

            return False

        def is_vwvh(value):
            if type(value) is str and (value.endswith("vw") or value.endswith("vh")):
                num = float(value[:-2])
                return num >= 0 and num <= 100

            return False

        if size is None:
            return None
        if is_non_negative_number(size):
            return number_to_str(size)
        elif is_percent(size) or is_vwvh(size):
            return size
        elif size == "fit-content":
            return size
        raise ValueError("Invalid size format.")


class ColorProp(Prop):
    def parse(self, color):
        return parse_color(color)


class BorderProp(Prop):
    """
    Accepts the following types:

    1. None
    2. Tuple with three values: (width in pixels , style , color)
    3. A dict that defines at least one of top, right, bottom, left
       where each entry is of type 1 or 2.
    If an entry is omitted from the dict, it's default value is None.
    """

    def parse(self, border):
        from .prop_utils import (
            parse_rectangle_style,
            is_color,
        )

        def is_border_style(style):
            return style in ["solid", "dotted", "dashed"]

        def border_to_str(border):
            from .theme import theme

            """
            Converts a tuple/list (width, style, color) into a single string
            that will be valid on the frontend.
            """
            if border is None:
                return None
            elif is_non_negative_number(border):
                thickness = border
                _, style, color = theme().input_border
                border = (thickness, style, color)
            elif is_color(border):
                color = parse_color(border)
                thickness, style, _ = theme().input_border
                border = (thickness, style, color)
            elif is_border_style(border):
                style = border
                thickness, _, color = theme().input_border
                border = (thickness, style, color)

            return f"{number_to_str(border[0])} {border[1]} {border[2]}"

        def is_border(border):
            """
            Returns True if the border is None
            or a tuple/list with three entries (width, style, color)
            """
            return (
                border is None
                or (
                    isinstance(border, (list, tuple))
                    and len(border) == 3
                    and is_non_negative_number(border[0])
                    and is_border_style(border[1])
                    and is_color(border[2])
                )
                or is_color(border)
                or is_non_negative_number(border)
                or is_border_style(border)
            )

        try:
            return parse_rectangle_style(border, is_border, border_to_str, default=None)
        except ValueError:
            raise ValueError("Invalid border format.")


class OutlineProp(BorderProp):
    def parse(self, outline):
        if outline is None or outline == "auto":
            return outline

        offset = 0
        # TODO: This is ambiguous, because a border could be
        # [1,1,1,1], which case this code takes the border as (1,1,1)
        # ans the offset as 1. It also does not do enough checking on
        # the args. The 4th arg could be a string, tuple, etc.
        if type(outline) in (list, tuple) and len(outline) == 4:
            offset = outline[3]
            outline = outline[:-1]

        return dict(outline=super().parse(outline), offset=number_to_str(offset))


class BorderRadiusProp(Prop):
    """
    Accepts the following types:

    1. A number (int or float)
    2. A dict that defines least one of topleft, topright, bottomleft, bottomright
       If an entry is omitted from the dict, it's default value is 0.
    """

    def parse(self, radius):
        from .prop_utils import parse_rectangle_style

        try:
            return parse_rectangle_style(
                radius,
                is_non_negative_number,
                lambda x: x,
                names=("topleft", "topright", "bottomright", "bottomleft"),
                default=0,
            )
        except ValueError:
            raise ValueError("Invalid border radius format.")


class PaddingProp(Prop):
    """
    Accepts the following types:

    1. A number (int or float) that represnts rem units
    2. A dict that defines at least one of top, right, bottom, left
       If an entry is omitted from the dict, it's default value is 0
    """

    def parse(self, padding):
        from .prop_utils import parse_rectangle_style

        try:
            return parse_rectangle_style(
                padding, is_non_negative_number, lambda x: x, default=0
            )
        except ValueError:
            raise ValueError("Invalid padding format.")


class CursorProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, (None, "auto", "pointer"), **kwargs)


class ThemeModeProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, ("system", "dark", "light"), **kwargs)


class SystemModeProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, ("dark", "light"), **kwargs)


alignments = ("start", "end", "center", "stretch")


class AlignProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, alignments, **kwargs)


directions = ("vertical", "horizontal")


class DirectionProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, directions, **kwargs)


justifies = (
    "start",
    "end",
    "center",
    "space-between",
    "space-around",
    "space-evenly",
    "stretch",
)


class JustifyProp(OptionProp):
    def __init__(self, value, **kwargs):
        super().__init__(value, justifies, **kwargs)


wraps = ("nowrap", "wrap", "reverse")


class WrapProp(Prop):
    def parse(self, wrap):
        if wrap is True:
            return "wrap"
        if wrap is False:
            return "nowrap"
        if wrap in wraps:
            return wrap
        raise ValueError("Invalid wrap value.")


class AriaProp(Prop):
    def parse(self, aria):
        if aria is None:
            return None
        if not isinstance(aria, dict):
            raise ValueError("aria must be a dict.")
        for k, v in aria.items():
            if not (isinstance(k, str) and k.isalpha()):
                raise ValueError("aria keys must be alphabetic strings.")
            if not isinstance(v, (str, int, float, bool)):
                raise ValueError("aria values must be primitive.")
        return aria


class GapProp(SizeProp):
    def parse(self, gap):
        if type(gap) not in (list, tuple):
            return super().parse(gap)
        else:
            return f"{super().parse(gap[0])} {super().parse(gap[1])}"


class GridRowProp(Prop):
    def parse(self, gridrow):
        if gridrow is None:
            return None
        if (
            type(gridrow) in (list, tuple)
            and len(gridrow) == 2
            and all([isinstance(x, int) for x in gridrow])
        ):
            return f"{gridrow[0]} / {gridrow[1]}"

        raise ValueError(
            "Invalid grid row / column property. Must be a tuple of length 2."
        )


class GridTemplateAreasProp(Prop):
    def parse(self, areas):
        if areas is None:
            return None
        # TODO: Should this be a list of strings, or convert the
        # return list to contain only strings?
        elif type(areas) in (tuple, list):
            return areas

        raise ValueError("Invalid grid_template_areas format.")


class GridTemplateRowsProp(Prop):
    def parse(self, rows):
        # TODO: Write tests for this code
        def is_fr(value):
            if type(value) is str and (value.endswith("fr")):
                num = float(value[:-2])
                return num >= 0 and num <= 100

            return False

        if rows is None:
            return None
        elif type(rows) in (tuple, list):
            rowstr = ""
            for i in rows:
                if type(i) is int:
                    if i < 0:
                        raise ValueError("Value must be a positive int.")
                elif i != "auto" and not is_fr(i):
                    raise ValueError("Invalid grid_template_rows / columns.")
                rowstr += f"{i} "

            return rowstr.strip()

        raise ValueError("Invalid grid_template_rows / columns.")


class SrcProp(Prop):
    def parse(self, src):
        import os
        import xxhash

        from ..core.utils import static_directory
        from .prop_utils import is_url

        if src is None or src == "":
            return src

        if is_url(src):
            return src

        src = os.path.abspath(src)
        if not os.path.exists(src):
            raise ValueError("File does not exist.")

        sourcedir = os.path.dirname(src)
        filename = os.path.basename(src)

        if sourcedir != static_directory:
            sourcedir_hash = xxhash.xxh32(sourcedir).hexdigest()
            filename = f"{sourcedir_hash}_{filename}"

        hosted_path = os.path.join(static_directory, filename)
        # Do not copy the file if it already
        # exists at the destination
        if not os.path.exists(hosted_path):
            os.symlink(src, hosted_path)

        return f"/static/{filename}"
